"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseHeaders = exports.getClientResponseDetail = exports.getClientRequestDetail = void 0;
var stream_1 = require("stream");
var utils_1 = require("../utils");
function getClientRequestDetail(request, options) {
    if (options === void 0) { options = {}; }
    return new Promise(function (resolve, reject) {
        var headers = parseHeaders(request.getHeaders());
        var req = {
            time: Date.now(),
            headers: headers,
            contentType: (0, utils_1.hasBody)(request.method) ? ((0, utils_1.contentType)(headers) || undefined) : undefined,
            method: request.method,
            url: "".concat(request.protocol, "//").concat(request.getHeader('host')).concat(request.path),
        };
        if (options.body !== true || !req.contentType) {
            resolve(req);
            return;
        }
        var input = new stream_1.PassThrough();
        var decoders = (0, utils_1.createDecoders)(headers);
        request.prependOnceListener('error', function (error) {
            reject(error);
            input.end();
        });
        var reqWrite = request.write;
        request.write = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            /**
             * chunk can be either a string or a Buffer.
             */
            var chunk = args[0];
            if (Buffer.isBuffer(chunk)) {
                input.push(chunk);
            }
            else {
                input.push(Buffer.from(chunk, 'utf8'));
            }
            return reqWrite.apply(this, args);
        };
        var reqEnd = request.end;
        request.end = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            /**
             * the first argument might be a callback or a chunk
             */
            var maybeChunk = args[0];
            if (Buffer.isBuffer(maybeChunk)) {
                input.push(maybeChunk);
            }
            else if (maybeChunk && typeof maybeChunk !== 'function') {
                input.push(Buffer.from(maybeChunk, 'utf8'));
            }
            input.end();
            return reqEnd.apply(this, args);
        };
        (0, utils_1.pipeline)(input, decoders)
            .then(function (buffer) {
            req.contentText = buffer === null || buffer === void 0 ? void 0 : buffer.toString('utf-8');
            resolve(req);
        })
            .catch(function (error) {
            reject(error);
        });
    });
}
exports.getClientRequestDetail = getClientRequestDetail;
function getClientResponseDetail(request, options) {
    if (options === void 0) { options = {}; }
    return new Promise(function (resolve, reject) {
        var start = Date.now();
        request.prependOnceListener('error', function (error) {
            reject(error);
        });
        request.prependOnceListener('response', function (response) {
            var _a;
            var now = Date.now();
            var headers = parseHeaders(response.headers);
            var res = {
                time: Date.now(),
                status: (_a = response.statusCode) !== null && _a !== void 0 ? _a : 0,
                headers: headers,
                contentType: (0, utils_1.contentType)(headers) || undefined,
                cost: now - start,
            };
            if (options.body !== true || !res.contentType) {
                resolve(res);
                return;
            }
            var input = new stream_1.PassThrough();
            var decoders = (0, utils_1.createDecoders)(headers);
            var aborted = false;
            var timeout = setTimeout(function () {
                input.end();
                timeout = null;
                aborted = true;
            }, 1000);
            response.on('error', function (error) {
                reject(error);
                input.end();
            });
            var push = response.push;
            response.push = function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                if (aborted === false) {
                    if (timeout !== null) {
                        clearTimeout(timeout);
                        timeout = null;
                    }
                    input.push.apply(input, args);
                }
                return push.apply(this, args);
            };
            response.prependOnceListener('end', function () {
                if (aborted === false) {
                    input.end();
                }
            });
            (0, utils_1.pipeline)(input, decoders)
                .then(function (buffer) {
                res.contentText = buffer === null || buffer === void 0 ? void 0 : buffer.toString('utf-8');
                resolve(res);
            })
                .catch(function (error) {
                reject(error);
            });
        });
    });
}
exports.getClientResponseDetail = getClientResponseDetail;
function parseHeaders(headerObj) {
    var headers = [];
    for (var _i = 0, _a = Object.keys(headerObj); _i < _a.length; _i++) {
        var rawName = _a[_i];
        var name_1 = rawName.toLowerCase();
        var rawValue = headerObj[rawName];
        if (Array.isArray(rawValue)) {
            for (var _b = 0, rawValue_1 = rawValue; _b < rawValue_1.length; _b++) {
                var value = rawValue_1[_b];
                headers.push([name_1, value !== undefined ? String(value) : '']);
            }
        }
        else {
            headers.push([name_1, rawValue !== undefined ? String(rawValue) : '']);
        }
    }
    return headers;
}
exports.parseHeaders = parseHeaders;
//# sourceMappingURL=http.js.map