"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseHeaders = exports.getServerResponseDetail = exports.getServerRequestDetail = void 0;
var stream_1 = require("stream");
var utils_1 = require("../utils");
function getServerRequestDetail(request, options) {
    if (options === void 0) { options = {}; }
    return new Promise(function (resolve, reject) {
        var headers = parseHeaders(request.headers);
        var req = {
            time: Date.now(),
            headers: headers,
            method: request.method,
            url: request.url,
            contentType: (0, utils_1.hasBody)(request.method) ? ((0, utils_1.contentType)(headers) || undefined) : undefined,
        };
        if (options.body !== true || !req.contentType) {
            resolve(req);
            return;
        }
        var input = new stream_1.PassThrough();
        var decoders = (0, utils_1.createDecoders)(headers);
        var aborted = false;
        var timeout = setTimeout(function () {
            input.end();
            timeout = null;
            aborted = true;
        }, 1000);
        request.on('error', function (error) {
            reject(error);
            input.end();
        });
        var push = request.push;
        request.push = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            if (aborted === false) {
                if (timeout !== null) {
                    clearTimeout(timeout);
                    timeout = null;
                }
                input.push.apply(input, args);
            }
            return push.apply(this, args);
        };
        request.prependOnceListener('end', function () {
            if (aborted === false) {
                input.end();
            }
        });
        (0, utils_1.pipeline)(input, decoders)
            .then(function (buffer) {
            req.contentText = buffer === null || buffer === void 0 ? void 0 : buffer.toString('utf-8');
            resolve(req);
        })
            .catch(function (error) {
            reject(error);
        });
    });
}
exports.getServerRequestDetail = getServerRequestDetail;
function getServerResponseDetail(response, options) {
    if (options === void 0) { options = {}; }
    return new Promise(function (resolve, reject) {
        var start = Date.now();
        var input = null;
        var decoders = [];
        var res = {
            time: start,
            status: 0,
            headers: [],
            cost: 0,
        };
        response.prependOnceListener('error', function (error) {
            reject(error);
            if (input) {
                input.end();
            }
        });
        var resolveHeader = undefined;
        var handleHeader = function () {
            if (resolveHeader !== undefined) {
                return resolveHeader;
            }
            var headers = parseHeaders(response.getHeaders());
            res.headers = headers;
            res.status = response.statusCode;
            res.contentType = (0, utils_1.contentType)(headers) || undefined;
            res.cost = Date.now() - start;
            if (options.body !== true || !res.contentType) {
                resolve(res);
                if (input) {
                    input.end();
                }
                return resolveHeader = true;
            }
            input = new stream_1.PassThrough();
            decoders = (0, utils_1.createDecoders)(headers);
            (0, utils_1.pipeline)(input, decoders)
                .then(function (buffer) {
                res.contentText = buffer === null || buffer === void 0 ? void 0 : buffer.toString('utf-8');
                resolve(res);
            })
                .catch(function (error) {
                reject(error);
            });
            return resolveHeader = false;
        };
        var resWrite = response.write;
        response.write = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            handleHeader();
            /**
             * chunk can be either a string or a Buffer.
             */
            var chunk = args[0];
            if (input) {
                if (Buffer.isBuffer(chunk)) {
                    input.push(chunk);
                }
                else {
                    input.push(Buffer.from(chunk, 'utf8'));
                }
            }
            return resWrite.apply(this, args);
        };
        var resEnd = response.end;
        response.end = function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            handleHeader();
            /**
             * the first argument might be a callback or a chunk
             */
            var maybeChunk = args[0];
            if (input) {
                if (Buffer.isBuffer(maybeChunk)) {
                    input.push(maybeChunk);
                }
                else if (maybeChunk && typeof maybeChunk !== 'function') {
                    input.push(Buffer.from(maybeChunk, 'utf8'));
                }
                input.end();
            }
            return resEnd.apply(this, args);
        };
    });
}
exports.getServerResponseDetail = getServerResponseDetail;
function parseHeaders(headerObj) {
    var headers = [];
    for (var _i = 0, _a = Object.keys(headerObj); _i < _a.length; _i++) {
        var rawName = _a[_i];
        var name_1 = rawName.toLowerCase();
        var rawValue = headerObj[rawName];
        if (Array.isArray(rawValue)) {
            for (var _b = 0, rawValue_1 = rawValue; _b < rawValue_1.length; _b++) {
                var value = rawValue_1[_b];
                headers.push([name_1, value !== undefined ? String(value) : '']);
            }
        }
        else {
            headers.push([name_1, rawValue !== undefined ? String(rawValue) : '']);
        }
    }
    return headers;
}
exports.parseHeaders = parseHeaders;
//# sourceMappingURL=http.js.map