import { ClientRequest } from 'http';
import { PassThrough } from 'stream';
import { ClientRequestDetail, ClientResponseDetail } from './types';
import { hasBody, contentType, createDecoders, pipeline } from '../utils';

export function getClientRequestDetail(request: ClientRequest, options: { body?: boolean } = {}): Promise<ClientRequestDetail> {
  return new Promise((resolve, reject) => {
    const headers = parseHeaders(request.getHeaders());
    const req: ClientRequestDetail = {
      time: Date.now(),
      headers: headers,
      contentType: hasBody(request.method) ? (contentType(headers) || undefined) : undefined,
      method: request.method,
      url: `${request.protocol}//${request.getHeader('host')}${request.path}`,
    };

    if (options.body !== true || !req.contentType) {
      resolve(req);
      return;
    }

    const input = new PassThrough();
    const decoders = createDecoders(headers);

    request.prependOnceListener('error', error => {
      reject(error);
      input.end();
    });

    const reqWrite = request.write;
    request.write = function (...args: any) {
      /**
       * chunk can be either a string or a Buffer.
       */
      const [chunk] = args;
      if (Buffer.isBuffer(chunk)) {
        input.push(chunk);
      } else {
        input.push(Buffer.from(chunk, 'utf8'));
      }

      return reqWrite.apply(this, args);
    };

    const reqEnd = request.end;
    request.end = function (...args: any) {
      /**
       * the first argument might be a callback or a chunk
       */
      const [maybeChunk] = args;

      if (Buffer.isBuffer(maybeChunk)) {
        input.push(maybeChunk);
      } else if (maybeChunk && typeof maybeChunk !== 'function') {
        input.push(Buffer.from(maybeChunk, 'utf8'));
      }
      input.end();

      return reqEnd.apply(this, args);
    };

    pipeline(input, decoders)
      .then(buffer => {
        req.contentText = buffer?.toString('utf-8');
        resolve(req);
      })
      .catch((error: Error) => {
        reject(error);
      })
  });
}

export function getClientResponseDetail(request: ClientRequest, options: { body?: boolean } = {}): Promise<ClientResponseDetail> {
  return new Promise((resolve, reject) => {
    const start = Date.now();
    request.prependOnceListener('error', (error) => {
      reject(error);
    });
    request.prependOnceListener('response', (response) => {
      const now = Date.now();
      const headers = parseHeaders(response.headers);
      const res: ClientResponseDetail = {
        time: Date.now(),
        status: response.statusCode ?? 0,
        headers: headers,
        contentType: contentType(headers) || undefined,
        cost: now - start,
      };
      if (options.body !== true || !res.contentType) {
        resolve(res);
        return;
      }
      const input = new PassThrough();
      const decoders = createDecoders(headers);
      let aborted = false;
      let timeout: any = setTimeout(() => {
        input.end();
        timeout = null;
        aborted = true;
      }, 1000);

      response.on('error', error => {
        reject(error);
        input.end();
      });

      const push = response.push;
      response.push = function(...args) {
        if (aborted === false) {
          if (timeout !== null) {
            clearTimeout(timeout);
            timeout = null;
          }
          input.push(...args);
        }
        return push.apply(this, args);
      }
      response.prependOnceListener('end', () => {
        if (aborted === false) {
          input.end();
        }
      });

      pipeline(input, decoders)
        .then(buffer => {
          res.contentText = buffer?.toString('utf-8');
          resolve(res);
        })
        .catch((error: Error) => {
          reject(error);
        })
    });

  });
}

export function parseHeaders(headerObj: Record<string, string[] | string | number | undefined>): [name: string, value: string][] {
  const headers: [name: string, value: string][] = [];
  for (const rawName of Object.keys(headerObj)) {
    const name = rawName.toLowerCase();
    const rawValue = headerObj[rawName];
    if (Array.isArray(rawValue)) {
      for (const value of rawValue) {
        headers.push([name, value !== undefined ? String(value) : '']);
      }
    } else {
      headers.push([name, rawValue !== undefined ? String(rawValue) : '']);
    }
  }
  return headers;
}