import diagnosticsChannel from 'diagnostics_channel';
import http, { ClientRequest } from 'http';
import https from 'https';
import semver from 'semver';
import { ClientHttpEvent } from './types';
import { getClientRequestDetail, getClientResponseDetail, parseHeaders } from './http';
import { getClientUndiciRequestDetail, getClientUndiciResponseDetail, UndiciRequset, parseUndiciRequsetHeaders } from './undici';
import { createListener, addSequence, errorToString } from '../utils';

const [ clientHttpInspector, emit ] = createListener<[ClientHttpEvent]>();

const httpRequestChannel = diagnosticsChannel.channel('http.client.request.start');
const undiciRequestChannel = diagnosticsChannel.channel('undici:request:create');

if (semver.satisfies(process.versions.node, '<=16')) {
  const inspectedClientMethod = function(
    this: typeof http | typeof https,
    func: (...args: any[]) => ClientRequest,
    ...rest: any[]
  ) {
    const request: ClientRequest = func.apply(this, rest);
    const sequence = addSequence();
    if (clientHttpInspector.hasListeners) {

      let callable = true;
      let requestDetailBody = false;
      const requestDetailResolvers: ((data: ReturnType<ClientHttpEvent['getRequestDetail']>) => void)[] = [];
      let responseDetailBody = false;
      const responseDetailResolvers: ((data: ReturnType<ClientHttpEvent['getResponseDetail']>) => void)[] = [];
      emit({
        url: `${request.protocol}//${request.getHeader('host')}${request.path}`,
        method: request.method,
        time: Date.now(),
        requestHeaders: parseHeaders(request.getHeaders()),
        getRequestDetail(options) {
          if (!callable) {
            return Promise.reject('event.getRequestDetail needs be called imminently');
          }
          if (requestDetailBody === false && options?.body === true) {
            requestDetailBody = true;
          }
          return new Promise(resolve => requestDetailResolvers.push(resolve));
        },
        getResponseDetail(options) {
          if (!callable) {
            return Promise.reject('event.getResponseDetail needs be called imminently');
          }
          if (responseDetailBody === false && options?.body === true) {
            responseDetailBody = true;
          }
          return new Promise(resolve => responseDetailResolvers.push(resolve));
        },
        sequence,
      });
      callable = false;
      if (requestDetailResolvers.length) {
        const promise = getClientRequestDetail(request, { body: requestDetailBody }).catch(error => ({ error: errorToString(error) }))
        for (const resolver of requestDetailResolvers) {
          resolver(promise);
        }
        requestDetailResolvers.length = 0;
      }
      if (responseDetailResolvers.length) {
        const promise = getClientResponseDetail(request, { body: responseDetailBody }).catch(error => ({ error: errorToString(error) }))
        for (const resolver of responseDetailResolvers) {
          resolver(promise);
        }
        responseDetailResolvers.length = 0;
      }
    }
    return request;
  };
  http.request = inspectedClientMethod.bind(http, http.request);
  http.get = inspectedClientMethod.bind(http, http.get);
  https.request = inspectedClientMethod.bind(https, https.request);
  https.get = inspectedClientMethod.bind(https, https.get);
} else {
  httpRequestChannel.subscribe(async (event: any) => {
    const request: ClientRequest = event.request;
    const sequence = addSequence();
    if (clientHttpInspector.hasListeners) {
      let callable = true;
      let requestDetailBody = false;
      const requestDetailResolvers: ((data: ReturnType<ClientHttpEvent['getRequestDetail']>) => void)[] = [];
      let responseDetailBody = false;
      const responseDetailResolvers: ((data: ReturnType<ClientHttpEvent['getResponseDetail']>) => void)[] = [];
      emit({
        url: `${request.protocol}//${request.getHeader('host')}${request.path}`,
        method: request.method,
        time: Date.now(),
        requestHeaders: parseHeaders(request.getHeaders()),
        getRequestDetail(options) {
          if (!callable) {
            return Promise.reject('event.getRequestDetail needs be called imminently');
          }
          if (requestDetailBody === false && options?.body === true) {
            requestDetailBody = true;
          }
          return new Promise(resolve => requestDetailResolvers.push(resolve));
        },
        getResponseDetail(options) {
          if (!callable) {
            return Promise.reject('event.getResponseDetail needs be called imminently');
          }
          if (responseDetailBody === false && options?.body === true) {
            responseDetailBody = true;
          }
          return new Promise(resolve => responseDetailResolvers.push(resolve));
        },
        sequence,
      });
      callable = false;
      if (requestDetailResolvers.length) {
        const promise = getClientRequestDetail(request, { body: requestDetailBody }).catch(error => ({ error: errorToString(error) }))
        for (const resolver of requestDetailResolvers) {
          resolver(promise);
        }
        requestDetailResolvers.length = 0;
      }
      if (responseDetailResolvers.length) {
        const promise = getClientResponseDetail(request, { body: responseDetailBody }).catch(error => ({ error: errorToString(error) }))
        for (const resolver of responseDetailResolvers) {
          resolver(promise);
        }
        responseDetailResolvers.length = 0;
      }
    }
  });
}

undiciRequestChannel.subscribe(async (event: any) => {
  const request: UndiciRequset = event.request;
  const sequence = addSequence();
  if (clientHttpInspector.hasListeners) {
    let callable = true;
    let requestDetailBody = false;
    const requestDetailResolvers: ((data: ReturnType<ClientHttpEvent['getRequestDetail']>) => void)[] = [];
    let responseDetailBody = false;
    const responseDetailResolvers: ((data: ReturnType<ClientHttpEvent['getResponseDetail']>) => void)[] = [];
    emit({
      url: (request as any).origin + request.path,
      method: request.method,
      time: Date.now(),
      requestHeaders: parseUndiciRequsetHeaders(request.headers),
      getRequestDetail(options) {
        if (!callable) {
          return Promise.reject('event.getRequestDetail needs be called imminently');
        }
        if (requestDetailBody === false && options?.body === true) {
          requestDetailBody = true;
        }
        return new Promise(resolve => requestDetailResolvers.push(resolve));
      },
      getResponseDetail(options) {
        if (!callable) {
          return Promise.reject('event.getResponseDetail needs be called imminently');
        }
        if (responseDetailBody === false && options?.body === true) {
          responseDetailBody = true;
        }
        return new Promise(resolve => responseDetailResolvers.push(resolve));
      },
      sequence,
    });
    callable = false;
    if (requestDetailResolvers.length) {
      const promise = getClientUndiciRequestDetail(request, { body: requestDetailBody }).catch(error => ({ error: errorToString(error) }))
      for (const resolver of requestDetailResolvers) {
        resolver(promise);
      }
      requestDetailResolvers.length = 0;
    }
    if (responseDetailResolvers.length) {
      const promise = getClientUndiciResponseDetail(request, { body: responseDetailBody }).catch(error => ({ error: errorToString(error) }))
      for (const resolver of responseDetailResolvers) {
        resolver(promise);
      }
      responseDetailResolvers.length = 0;
    }
  }
});

export default clientHttpInspector;