import { PassThrough, Transform } from 'stream';
import { ClientRequestDetail, ClientResponseDetail } from './types';
import { hasBody, contentType, createDecoders, pipeline } from '../utils';

export interface UndiciRequset {
  path: string;
  // 'name1: value\r\nname2: value2\r\n';
  headers: string;
  // GET PUST
  method: string;
  onBodySent(chunk: Buffer): void;
  onRequestSent(): void;
  onError(error: Error): void;
  onHeaders(statusCode: number, headers: Buffer[] | string[] | null, resume: () => void): boolean;
  onData(chunk: Buffer): boolean;
  onComplete(trailers: string[] | null): void;
}

export function getClientUndiciRequestDetail(request: UndiciRequset, options: { body?: boolean } = {}): Promise<ClientRequestDetail> {

  return new Promise((resolve, reject) => {
    const headers = parseUndiciRequsetHeaders(request.headers);
    const req: ClientRequestDetail = {
      method: request.method,
      url: (request as any).origin + request.path,
      time: Date.now(),
      headers: headers,
      contentType: hasBody(request.method) ? (contentType(headers) || undefined) : undefined,
    };
    if (options.body !== true || !req.contentType) {
      resolve(req);
      return;
    }
    const input = new PassThrough();
    const decoders = createDecoders(headers);

    const onError = request.onError;
    request.onError = function(this: Request, error: Error) {
      reject(error);
      input.end();
      return onError.call(this, error);
    }
    const onBodySent = request.onBodySent;
    request.onBodySent = function(this: Request, chunk: Buffer) {
      input.push(chunk);
      return onBodySent.call(this, chunk);
    }
    const onRequestSent = request.onRequestSent;
    request.onRequestSent = function(this: Request) {
      input.end();
      return onRequestSent.call(this);
    }

    pipeline(input, decoders)
      .then(buffer => {
        req.contentText = buffer?.toString('utf-8');
        resolve(req);
      })
      .catch((error: Error) => {
        reject(error);
      })
  });
}

export function getClientUndiciResponseDetail(request: UndiciRequset, options: { body?: boolean } = {}): Promise<ClientResponseDetail> {
  return new Promise((resolve, reject) => {
    const res: ClientResponseDetail = {
      time: 0,
      status: 0,
      cost: 0,
      headers: [],
      contentType: undefined,
    };

    let input: PassThrough | null = null;
    let decoders: Transform[] = [];

    const onHeaders = request.onHeaders;

    request.onHeaders = function(status, headersList, resume) {
      const headers = parseUndiciResponseHeaders(headersList);

      res.status = status;
      res.headers = headers;
      res.time = Date.now();
      res.contentType = contentType(headers) || undefined

      if (options.body !== true || !res.contentType) {
        resolve(res);
      } else {
        input = new PassThrough();
        decoders = createDecoders(headers);
        pipeline(input, decoders)
          .then(buffer => {
            res.contentText = buffer?.toString('utf-8');
            resolve(res);
          })
          .catch((error: Error) => {
            reject(error);
          })
      }
      return onHeaders.call(this, status, headersList, resume);
    }

    const onData = request.onData;
    request.onData = function(chunk) {
      if (input) {
        input.push(chunk);
      }
      return onData.call(this, chunk);
    }

    const onComplete = request.onComplete;
    request.onComplete = function(trailers) {
      if (input) {
        input.end();
      }
      return onComplete.call(this, trailers);
    }
  });
}

export function parseUndiciRequsetHeaders(headerString: string): [name: string, value: string][] {
  const result: [name: string, value: string][] = [];
  const segments = headerString.trim().split(/[\n\r\t\f\v]+/);
  for (const segment of segments) {
    const [_name, _value] = segment.split(/\s*\:\s*/);
    const name = _name.toLowerCase();
    const value = _value ?? '';
    if (!name) {
      continue;
    }
    result.push([name, value]);
  }
  return result;
}

export function parseUndiciResponseHeaders(headersList: string[] | Buffer[] | null): [name: string, value: string][] {
  const result: [name: string, value: string][] = [];
  if (headersList) {
    for (let n = 0; n < headersList.length; n += 2) {
      const name = headersList[n + 0].toString('latin1').toLowerCase();
      const value = headersList[n + 1].toString('latin1');
      result.push([name, value]);
    }
  }
  return result;
}