import { IncomingMessage, ServerResponse } from 'http';
import { PassThrough, Transform } from 'stream';
import { ServerRequestDetail, ServerResponseDetail } from './types';
import { hasBody, contentType, createDecoders, pipeline } from '../utils';

export function getServerRequestDetail(request: IncomingMessage, options: { body?: boolean } = {}): Promise<ServerRequestDetail> {
  return new Promise((resolve, reject) => {
    const headers = parseHeaders(request.headers);
    const req: ServerRequestDetail = {
      time: Date.now(),
      headers: headers,
      method: request.method!,
      url: request.url!,
      contentType: hasBody(request.method!) ? (contentType(headers) || undefined) : undefined,
    }

    if (options.body !== true || !req.contentType) {
      resolve(req);
      return;
    }

    const input = new PassThrough();
    const decoders = createDecoders(headers);
    let aborted = false;
    let timeout: any = setTimeout(() => {
      input.end();
      timeout = null;
      aborted = true;
    }, 1000);

    request.on('error', error => {
      reject(error);
      input.end();
    });

    const push = request.push;
    request.push = function(...args) {
      if (aborted === false) {
        if (timeout !== null) {
          clearTimeout(timeout);
          timeout = null;
        }
        input.push(...args);
      }
      return push.apply(this, args);
    }
    request.prependOnceListener('end', () => {
      if (aborted === false) {
        input.end();
      }
    });

    pipeline(input, decoders)
      .then(buffer => {
        req.contentText = buffer?.toString('utf-8');
        resolve(req);
      })
      .catch((error: Error) => {
        reject(error);
      })
  });
}

export function getServerResponseDetail(response: ServerResponse, options: { body?: boolean } = {}): Promise<ServerResponseDetail> {
  return new Promise((resolve, reject) => {
    const start = Date.now();

    let input: PassThrough | null = null;
    let decoders: Transform[] = [];

    const res: ServerResponseDetail = {
      time: start,
      status: 0,
      headers: [],
      cost: 0,
    };

    response.prependOnceListener('error', error => {
      reject(error);
      if (input) {
        input.end();
      }
    });

    let resolveHeader: boolean | undefined = undefined;
    const handleHeader = () => {
      if (resolveHeader !== undefined) {
        return resolveHeader;
      }
      const headers = parseHeaders(response.getHeaders());
      res.headers = headers;
      res.status = response.statusCode;
      res.contentType = contentType(headers) || undefined;
      res.cost = Date.now() - start;
      if (options.body !== true || !res.contentType) {
        resolve(res);
        if (input) {
          input.end();
        }
        return resolveHeader = true;
      }
      input = new PassThrough();
      decoders = createDecoders(headers);
      pipeline(input, decoders)
      .then(buffer => {
        res.contentText = buffer?.toString('utf-8');
        resolve(res);
      })
      .catch((error: Error) => {
        reject(error);
      })
      return resolveHeader = false;
    }

    const resWrite = response.write;
    response.write = function (...args: any) {
      handleHeader();
      /**
       * chunk can be either a string or a Buffer.
       */
      const [chunk] = args;
      if (input) {
        if (Buffer.isBuffer(chunk)) {
          input.push(chunk);
        } else {
          input.push(Buffer.from(chunk, 'utf8'));
        }
      }

      return resWrite.apply(this, args);
    };

    const resEnd = response.end;
    response.end = function (...args: any) {
      handleHeader();
      /**
       * the first argument might be a callback or a chunk
       */
      const [maybeChunk] = args;

      if (input) {
        if (Buffer.isBuffer(maybeChunk)) {
          input.push(maybeChunk);
        } else if (maybeChunk && typeof maybeChunk !== 'function') {
          input.push(Buffer.from(maybeChunk, 'utf8'));
        }
        input.end();
      }

      return resEnd.apply(this, args);
    };
  });
}

export function parseHeaders(headerObj: Record<string, string[] | string | number | undefined>): [name: string, value: string][] {
  const headers: [name: string, value: string][] = [];
  for (const rawName of Object.keys(headerObj)) {
    const name = rawName.toLowerCase();
    const rawValue = headerObj[rawName];
    if (Array.isArray(rawValue)) {
      for (const value of rawValue) {
        headers.push([name, value !== undefined ? String(value) : '']);
      }
    } else {
      headers.push([name, rawValue !== undefined ? String(rawValue) : '']);
    }
  }
  return headers;
}