
import http, { Server, IncomingMessage, ServerResponse } from 'http';
import https from 'https';
import { createListener, startSequence, addSequence, errorToString } from '../utils';
import { ServerHttpEvent } from './types';
import { getServerRequestDetail, getServerResponseDetail, parseHeaders } from './http';

const [ serverHttpInspector, emit ] = createListener<[ServerHttpEvent]>();

const inspectedServerMethod = function(
  this: typeof http | typeof https,
  func: (...args: any[]) => Server,
  ...rest: any[]
): any {
  const originHandler = typeof rest[rest.length - 1] === 'function' ? rest[rest.length - 1] : null;
  const handler = function(this: any, request: IncomingMessage, response: ServerResponse) {
    startSequence(sequence => {
      addSequence(sequence);
      if (serverHttpInspector.hasListeners) {
        let callable = true;
        let requestDetailBody = false;
        const requestDetailResolvers: ((data: ReturnType<ServerHttpEvent['getRequestDetail']>) => void)[] = [];
        let responseDetailBody = false;
        const responseDetailResolvers: ((data: ReturnType<ServerHttpEvent['getResponseDetail']>) => void)[] = [];
        emit({
          url: request.url!,
          method: request.method!,
          time: Date.now(),
          socket: {
            remoteAddress: request.socket.remoteAddress ?? '',
            encrypted: (request.socket as any).encrypted === true,
          },
          httpVersion: request.httpVersion,
          httpVersionMajor: request.httpVersionMajor,
          httpVersionMinor: request.httpVersionMinor,
          requestHeaders: parseHeaders(request.headers),
          getRequestDetail(options) {
            if (!callable) {
              return Promise.reject('event.getRequestDetail needs be called imminently');
            }
            if (requestDetailBody === false && options?.body === true) {
              requestDetailBody = true;
            }
            return new Promise(resolve => requestDetailResolvers.push(resolve));
          },
          getResponseDetail(options) {
            if (!callable) {
              return Promise.reject('event.getResponseDetail needs be called imminently');
            }
            if (responseDetailBody === false && options?.body === true) {
              responseDetailBody = true;
            }
            return new Promise(resolve => responseDetailResolvers.push(resolve));
          },
          sequence,
        });
        callable = false;
        if (requestDetailResolvers.length) {
          const promise = getServerRequestDetail(request, { body: requestDetailBody }).catch(error => ({ error: errorToString(error) }))
          for (const resolver of requestDetailResolvers) {
            resolver(promise);
          }
          requestDetailResolvers.length = 0;
        }
        if (responseDetailResolvers.length) {
          const promise = getServerResponseDetail(response, { body: responseDetailBody }).catch(error => ({ error: errorToString(error) }))
          for (const resolver of responseDetailResolvers) {
            resolver(promise);
          }
          responseDetailResolvers.length = 0;
        }
      }
      if (originHandler) {
        originHandler.call(this, request, response);
      }
    });
  }
  if (originHandler) {
    rest[rest.length - 1] = handler;
  } else {
    rest.push(handler);
  }
  return func.apply(this, rest);
};

http.createServer = inspectedServerMethod.bind(http, http.createServer);
https.createServer = inspectedServerMethod.bind(https, https.createServer);

export default serverHttpInspector;