import typeis from 'type-is';
import { Transform, Writable } from 'stream';
import { AsyncLocalStorage } from 'async_hooks';
import destroy from 'destroy';
import zlib from 'zlib';

function getHeader(name: string, headers: [name: string, value: string][]): string | null {
  name = name.toLowerCase();
  for (let header of headers) {
    if (header[0] === name) {
      return header[1];
    }
  }
  return null;
}

export function hasBody(method: string) {
  return ['POST', 'PUT', 'PATCH', 'DELETE'].includes(method.toUpperCase());
}

export function contentType(headers: [name: string, value: string][]) {
  const contentType = getHeader('content-type', headers) ?? '';
  return typeis.is(contentType, [ 'urlencoded', 'json', 'text', 'html' ]);
}

const zlibOptions = {
  flush: zlib.constants.Z_SYNC_FLUSH,
  finishFlush: zlib.constants.Z_SYNC_FLUSH
};

const brotliOptions = {
  flush: zlib.constants.BROTLI_OPERATION_FLUSH,
  finishFlush: zlib.constants.BROTLI_OPERATION_FLUSH
};

export function createDecoders(headers: [name: string, value: string][]): Transform[] {
  const decoders: Transform[] = [];
  const encodings = (getHeader('content-encoding', headers) ?? '')
    .split(',')
    .map(val => val.trim().toLowerCase())
    .filter(val => val.length);

  for (const encoding of encodings) {
    switch (encoding) {
      case 'gzip':
      case 'x-gzip':
      case 'compress':
      case 'x-compress':
      case 'deflate':
        decoders.push(zlib.createUnzip(zlibOptions));
        break;
      case 'br':
        decoders.push(zlib.createBrotliDecompress(brotliOptions));
        break;
    }
  }
  return decoders;
}

export function pipeline(input: Writable, transforms: Transform[]): Promise<Buffer | null> {
  return new Promise((resolve, reject) => {
    for (let i = 0; i < transforms.length; i++) {
      (transforms[i - 1] ?? input).pipe(transforms[i]);
    }
    const output = transforms.length ? transforms[transforms.length - 1] : input;
    const chunks: Buffer[] = [];
    output.on('data', (chunk) => {
      chunks.push(chunk);
    });
    output.on('end', () => {
      destroy(input);
      for (const decoder of transforms) {
        destroy(decoder);
      }
      resolve(Buffer.concat(chunks));
    });
    output.on('error', error => {
      destroy(input);
      for (const decoder of transforms) {
        destroy(decoder);
      }
      reject(error);
    });
  });
}

export interface Listener<T extends any[]> {
  hasListeners: boolean;
  addListener(fn: (...args: T) => void): () => void;
  removeAllListeners(): void;
}

export interface Emit<T extends any[]> {
  (...args: T): void;
}

export function createListener<T extends any[]>(): [Listener<T>, Emit<T>] {
  const listeners: ((...args: T) => void)[] = [];
  const listener: Listener<T> = {
    get hasListeners() {
      return listeners.length > 0;
    },
    addListener(fn) {
      listeners.push(fn);
      return () => {
        const index = listeners.indexOf(fn);
        if (index >= 0) {
          listeners.splice(index, 1);
        }
      };
    },
    removeAllListeners() {
      listeners.length = 0;
    }
  };
  const emit: Emit<T> = (...args) => {
    for (const listener of listeners) {
      listener(...args);
    }
  };
  return [listener, emit];
}

const rand = Math.random;
const hexmap = [0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66];
const arr = new Array(36);

export function uuid() {
  setChars12(arr, rand(), 0);
  setChars12(arr, rand(), 12);
  setChars12(arr, rand(), 24);
  arr[8]  = 0x2d; // -
  arr[13] = 0x2d; // -
  arr[14] = 0x34; // 4
  arr[18] = 0x2d; // -
  arr[19] = hexmap[8 + (arr[19] & 3)]; // [89ab]
  arr[23] = 0x2d; // -
  return String.fromCharCode(...arr);
}

function setChars12(buf: number[], n: number, pos: number) {
  for (let i= 0; i < 12; i += 3) {
    n *= 0x1000;
    buf[pos + i + 0] = hexmap[(n      ) & 0xF];
    buf[pos + i + 1] = hexmap[(n >>> 4) & 0xF];
    buf[pos + i + 2] = hexmap[(n >>> 8) & 0xF];
  }
}

type SequenceData = { id: string, index: number };
const sequenceContext = new AsyncLocalStorage<SequenceData>();

export function startSequence(callback: (store: SequenceData) => void) {
  const store = { id: uuid(), index: -1 };
  sequenceContext.run(store, () => callback(store));
}

export function addSequence(sequenceData?: SequenceData): null | SequenceData {
  const store = sequenceData ?? sequenceContext.getStore();
  if (store) {
    store.index++;
    return store;
  }
  return null;
}

export function errorToString(error: unknown): string {
  if (error instanceof Error) {
    return error.message;
  }
  return String(error);
}